@@ -1,18 +1,7 @@
 /*
-    Copyright (C) 2018 Volker Krause <vkrause@kde.org>
+    SPDX-FileCopyrightText: 2018 Volker Krause <vkrause@kde.org>
 
-    This program is free software; you can redistribute it and/or modify it
-    under the terms of the GNU Library General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or (at your
-    option) any later version.
-
-    This program is distributed in the hope that it will be useful, but WITHOUT
-    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
-    License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+    SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 #include "timezonedbgenerator.h"
@@ -32,7 +21,7 @@ void TimezoneDbGenerator::generate(QIODevice *out)
 
     out->write(R"(
 #include "timezonedb_p.h"
-#include "timezonedb_data_p.h"
+#include "timezonedb_data.h"
 
 namespace KItinerary {
 namespace KnowledgeDb {
@@ -40,7 +29,7 @@ namespace KnowledgeDb {
 // timezone name strings
 static const char timezone_names[] =
 )");
-    // timezone string tables
+    // timezone string table
     for (const auto &tz : tzDb.m_zones) {
         out->write("    ");
         out->write("\"");
@@ -49,6 +38,22 @@ static const char timezone_names[] =
     }
     out->write(R"(;
 
+static constexpr const uint16_t timezone_names_offsets[] = {
+    )");
+    out->write(QByteArray::number(tzDb.m_zones.front().size()));
+    out->write(", // Undefined\n");
+
+    // offsets into timezone string table
+    for (const auto &tz : tzDb.m_zones) {
+        out->write("    ");
+        out->write(QByteArray::number(tzDb.offset(tz)));
+        out->write(", // ");
+        out->write(tz);
+        out->write("\n");
+    }
+
+    out->write(R"(};
+
 static constexpr const CountryTimezoneMap country_timezone_map[] = {
 )");
 
@@ -65,6 +70,20 @@ static constexpr const CountryTimezoneMap country_timezone_map[] = {
 
     out->write(R"(};
 
+static constexpr const CountryId timezone_country_map[] = {
+    CountryId{}, // Undefined
+)");
+
+    for (const auto &tz : tzDb.m_zones) {
+        out->write("    ");
+        CodeGen::writeCountryIsoCode(out, tzDb.m_countryForZone[tz]);
+        out->write(", // ");
+        out->write(tz);
+        out->write("\n");
+    }
+
+    out->write(R"(};
+
 }
 }
 )");
@@ -76,23 +95,22 @@ void TimezoneDbGenerator::generateHeader(QIODevice *out)
 
     Timezones tzDb;
     out->write(R"(
-#ifndef KITINERARY_KNOWLEDGEDB_TIMEZONEDB_DATA_P_H
-#define KITINERARY_KNOWLEDGEDB_TIMEZONEDB_DATA_P_H
+#ifndef KITINERARY_KNOWLEDGEDB_TIMEZONEDB_DATA_H
+#define KITINERARY_KNOWLEDGEDB_TIMEZONEDB_DATA_H
 
 #include <cstdint>
 
 namespace KItinerary {
 namespace KnowledgeDb {
 
-/** Enum representing all timezones, values match the offsets into the timezone name string table. */
+/** Enum representing all timezones. */
 enum class Tz : uint16_t {
+    Undefined,
 )");
 
     for (const auto &tz : tzDb.m_zones) {
         out->write("    ");
         CodeGen::writeTimezoneEnum(out, tz);
-        out->write(" = ");
-        out->write(QByteArray::number(tzDb.offset(tz)));
         out->write(",\n");
     }
     out->write(R"(};
